// ignore_for_file: implementation_imports, unnecessary_import
import 'package:analyzer/dart/element/element.dart';
import 'package:analyzer/src/dart/element/element.dart';
import 'package:build/build.dart';
import 'package:build/src/builder/build_step.dart';
import 'package:initializer_annotation/initializer_annotation.dart';
import 'package:path/path.dart' as p;
import 'package:source_gen/source_gen.dart';

import 'aggregate_generator_for_annotation.dart';
import 'aggregate_results.dart';
import 'settings.dart';

class InitializerAggregateResults extends AggregateResults {
  final List<TopLevelVariableElementImpl> _topLevelVariable = [];
  final List<FunctionElementImpl> _initFunction = [];
  final List<FieldElementImpl> _initField = [];
  final List<MethodElementImpl> _initMethod = [];

  final String groupName;
  @override
  final String path;
  InitializerAggregateResults({
    required this.groupName,
    required this.path,
  });
  bool get isAsync =>
      _initFunction.any((fun) => fun.isAsynchronous) ||
      _initMethod.any((fun) => fun.isAsynchronous);

  @override
  String toString() {
    final strbuf = StringBuffer() //
      ..writeln(
          '${isAsync ? 'Future<void>' : 'void'} ${groupName}Initializer() ${isAsync ? 'async ' : ''}{');
    if (_topLevelVariable.isNotEmpty) {
      for (final variable in _topLevelVariable) {
        strbuf.writeln(
            '  ${variable.getter?.isAsynchronous == true ? 'await ' : ''}${variable.identifier};');
      }
    }
    if (_initFunction.isNotEmpty) {
      for (final fn in _initFunction) {
        strbuf.writeln(
            '  ${fn.isAsynchronous ? 'await ' : ''}${fn.identifier}();');
      }
    }
    if (_initField.isNotEmpty) {
      for (final f in _initField) {
        final klass = f.enclosingElement as ClassElementImpl;
        strbuf.writeln(
            '  ${f.getter?.isAsynchronous == true ? 'await ' : ''}${klass.identifier}.${f.identifier};');
      }
    }
    if (_initMethod.isNotEmpty) {
      for (final m in _initMethod) {
        final klass = m.enclosingElement as ClassElementImpl;
        strbuf.writeln(
            '  ${m.isAsynchronous ? 'await ' : ''}${klass.identifier}.${m.identifier}();');
      }
    }
    strbuf.writeln('}');
    return strbuf.toString();
  }

  @override
  void merge(InitializerAggregateResults other) {
    _topLevelVariable.addAll(other._topLevelVariable);
    _initFunction.addAll(other._initFunction);
    _initField.addAll(other._initField);
    _initMethod.addAll(other._initMethod);
    super.merge(other);
  }

  void addTopLevelVariable(TopLevelVariableElementImpl e) {
    _topLevelVariable.add(e);
    addSourceMap(e.source.uri, e.identifier);
  }

  void addFunction(FunctionElementImpl e) {
    _initFunction.add(e);
    addSourceMap(e.source.uri, e.identifier);
  }

  void addField(FieldElementImpl e) {
    _initField.add(e);
    final klass = e.enclosingElement as ClassElementImpl;
    addSourceMap(e.librarySource.uri, klass.identifier);
  }

  void addMethod(MethodElementImpl e) {
    _initMethod.add(e);
    final klass = e.enclosingElement as ClassElementImpl;
    addSourceMap(e.librarySource.uri, klass.identifier);
  }
}

class InitializerAggregateResultsGroup
    extends AggregateResultsGroup<InitializerAggregateResults> {
  InitializerAggregateResultsGroup({
    super.groups,
    required super.rootPackage,
    required super.outputPath,
    required super.order,
  });
  @override
  String toString() {
    final strbuf = StringBuffer();
    final groupNames = <String>[];

    final srcMap = mergeSourceMap();
    for (final pair in srcMap.entries) {
      final src = pair.key;
      final nameList = pair.value;
      String? importPath;
      String? srcPath;
      switch (src.scheme) {
        case 'asset':
          srcPath = src.path;
          break;
        case 'package':
          srcPath = src.path.replaceFirst(rootPackage, '$rootPackage/lib');
          break;
      }
      final relativePath = p.dirname(p.join(rootPackage, outputPath));

      importPath = p
          .relative(
            srcPath!,
            from: relativePath,
          )
          .replaceAll(r'\', r'/');

      strbuf.writeln("import '$importPath' show ${nameList.join(', ')};");
    }
    strbuf.writeln();
    final asyncGroupMap = <String, bool>{};
    for (final kv in groups.entries) {
      final groupName = kv.key;
      final aggregateResults = kv.value;

      strbuf.writeln(aggregateResults.toString());
      groupNames.add(groupName);
      asyncGroupMap[groupName] = aggregateResults.isAsync;
    }
    final isAsync = asyncGroupMap.values.any((val) => val);
    strbuf.write('${isAsync ? 'Future<void>' : 'void'} initializer(');
    if (groupNames.isNotEmpty) strbuf.writeln('{');
    for (final groupName in groupNames) {
      strbuf.writeln('  bool enable$groupName = true,');
    }
    if (groupNames.isNotEmpty) strbuf.write('}');
    strbuf.writeln(')${isAsync ? ' async' : ''} {');
    if (order != null && order!.isNotEmpty) {
      final groupNameOrder = order!.toList();
      groupNames.sort(
        (a, b) {
          const intMax = 0x7fffffffffffffff;
          var aIndex = groupNameOrder.indexOf(a);
          if (aIndex < 0) aIndex = intMax;
          var bIndex = groupNameOrder.indexOf(b);
          if (bIndex < 0) bIndex = intMax;
          return aIndex.compareTo(bIndex);
        },
      );
    }
    for (final groupName in groupNames) {
      strbuf.writeln(
          '  if (enable$groupName) ${asyncGroupMap[groupName]! ? 'await ' : ''}${groupName}Initializer();');
    }
    strbuf.writeln('}');

    return strbuf.toString();
  }
}

class InitializerGenerator extends AggregateGeneratorForAnnotation<Initializer,
    InitializerAggregateResults> {
  final Settings _settings;

  Initializer get config => _settings.config.toInitializer();

  const InitializerGenerator.fromSettings(this._settings);

  factory InitializerGenerator({
    Initializer? config,
  }) =>
      InitializerGenerator.fromSettings(
        Settings(config: config),
      );

  @override
  MapEntry<String, InitializerAggregateResults> generateForAnnotatedElement(
    Element element,
    ConstantReader annotation,
    BuildStep buildStep,
  ) {
    final group = annotation.read('group');
    final groupString = group.isNull ? 'default' : group.stringValue;
    final result = InitializerAggregateResults(
      groupName: groupString,
      path: element.source!.fullName,
    );
    if (element is TopLevelVariableElementImpl) {
      result.addTopLevelVariable(element);
    } else if (element is PropertyAccessorElementImpl && element.isGetter) {
      final variable = element.variable;
      if (variable is TopLevelVariableElementImpl) {
        result.addTopLevelVariable(variable);
      }
    } else if (element is FunctionElementImpl) {
      result.addFunction(element);
    } else if (element is ClassElementImpl) {
      final klass = element;
      final annotatedFields =
          klass.fields.where((element) => typeChecker.hasAnnotationOf(element));
      for (final field in annotatedFields) {
        if (field.isStatic) {
          result.addField(field);
        } else {
          throw InvalidGenerationSourceError(
            '`@Initializer` can only be used on STATIC field.',
            element: element,
          );
        }
      }
      final annotatedMethods = klass.methods
          .where((element) => typeChecker.hasAnnotationOf(element));
      for (final method in annotatedMethods) {
        if (method.isStatic) {
          result.addMethod(method);
        } else {
          throw InvalidGenerationSourceError(
            '`@Initializer` can only be used on STATIC method.',
            element: element,
          );
        }
      }

      final annotatedAccessors = klass.accessors.where((element) =>
          typeChecker.hasAnnotationOf(element) && element.isGetter);

      for (final getter in annotatedAccessors) {
        final variable = getter.variable;
        if (getter.isStatic && variable is FieldElementImpl) {
          result.addField(variable);
        }
      }
    } else {
      const supported = [
        'top-level variable',
        'function without required args',
        'static member of class'
      ];
      throw InvalidGenerationSourceError(
        '`@Initializer` can only be used on $supported.',
        element: element,
      );
    }

    return MapEntry<String, InitializerAggregateResults>(groupString, result);
  }
}
